//////////////////////////////////////////////////////////////////////////
// uniformenu
// ==========
//
// Written by Tom Seddon <uniformenu@tomseddon.plus.com>
// 
//	Date		What
//	==================================================================
//	7/6/2003	first version
//	14/7/2003	adds 'close' item where appropriate rather than
//				indiscriminately resetting system menus. (that renders
//				console windows a bit useless...)
//	
#define STRICT
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <windowsx.h>
#include <psapi.h>
#include <shellapi.h>
#include <malloc.h>
#include <string.h>
#include "resource.h"

static const char *class_name="UniformenuClass";
static const char *mutex_name="UniformenuMutex";

static HWND main_window=0;

//A top-level window.
typedef struct {
	HWND handle;
	int fixed;
	int present;
} window_t;

enum {
	//increment for windows storage. not very important.
	max_num_windows_inc=512,
		
	//Message for tray notifications.
	tray_notification_msg=WM_APP+1,
	
	//id for tray icon
	tray_icon_id=0,
	
	//id for timer
	timer_id=1,
};

static unsigned max_num_windows=0;
static unsigned num_windows=0;
static window_t *windows=0;

static window_t *AddWindow(HWND h) {
	window_t *w;

	if(num_windows==max_num_windows) {
		max_num_windows+=max_num_windows_inc;
		windows=realloc(windows,max_num_windows*sizeof *windows);
	}

	w=&windows[num_windows++];
	w->handle=h;
	w->fixed=0;
	w->present=1;

	return w;
}

static window_t *WindowInfoFromHandle(HWND h) {
	unsigned i;

	for(i=0;i<num_windows;++i) {
		if(windows[i].handle==h) {
			return &windows[i];
		}
	}
	return AddWindow(h);
}

static BOOL CALLBACK EnumWindowProc(HWND hWnd,LPARAM lParam) {
	window_t *window;

	window=WindowInfoFromHandle(hWnd);
	if(!window->fixed&&IsWindowVisible(window->handle)) {
		HMENU menu;

		window->fixed=1;
		menu=GetSystemMenu(window->handle,FALSE);
		if(menu) {
			int item_count;

			item_count=GetMenuItemCount(menu);
			if(item_count>0&&GetMenuItemID(menu,item_count-1)!=SC_CLOSE) {
				AppendMenu(menu,MF_SEPARATOR,-1,"-");
				AppendMenu(menu,MF_STRING,SC_CLOSE,"&Close");
			}
		}
	}
	
	window->present=1;
	return TRUE;
}

static void FixMenus() {
	unsigned i,j;

	for(i=0;i<num_windows;++i) {
		windows[i].present=0;
	}
	EnumWindows(&EnumWindowProc,0);
	for(i=j=0;i<num_windows;++i) {
		windows[j]=windows[i];
		if(windows[j].present) {
			++j;
		}
	}
	num_windows=j;
}

static void AddTrayIcon() {
	NOTIFYICONDATA nid;

	nid.cbSize=sizeof nid;
	nid.hWnd=main_window;
	nid.uID=tray_icon_id;
	nid.uFlags=NIF_ICON|NIF_MESSAGE|NIF_TIP;
	nid.uCallbackMessage=tray_notification_msg;
	nid.hIcon=LoadIcon(GetModuleHandle(0),MAKEINTRESOURCE(IDI_ICON1));
	strcpy(nid.szTip,"uniformenu");
	Shell_NotifyIcon(NIM_ADD,&nid);
}

static void RemoveTrayIcon() {
	NOTIFYICONDATA nid;

	nid.cbSize=sizeof nid;
	nid.hWnd=main_window;
	nid.uID=tray_icon_id;
	Shell_NotifyIcon(NIM_DELETE,&nid);
}

static LRESULT CALLBACK WndProc(HWND h,UINT msg,WPARAM w,LPARAM l) {
	switch(msg) {
	case WM_TIMER:
		FixMenus();
		return 0;
	case tray_notification_msg:
		switch(l) {
		case WM_RBUTTONDOWN:
			PostQuitMessage(0);
			return 0;
		}
		return 0;
	case WM_CLOSE:
		PostQuitMessage(0);
		return 0;
	}
	return DefWindowProc(h,msg,w,l);
}

static void RegisterWindowClass() {
	WNDCLASSEX w;

	w.cbClsExtra=0;
	w.cbSize=sizeof w;
	w.cbWndExtra=0;
#ifdef _DEBUG
	w.hbrBackground=GetStockBrush(NULL_BRUSH);
	w.hCursor=LoadCursor(0,IDC_ARROW);
	w.hIconSm=LoadIcon(GetModuleHandle(0),MAKEINTRESOURCE(IDI_ICON1));
	w.hIcon=LoadIcon(GetModuleHandle(0),MAKEINTRESOURCE(IDI_ICON1));
#else
	w.hbrBackground=0;
	w.hCursor=0;
	w.hIconSm=0;
	w.hIcon=0;
#endif
	w.hInstance=GetModuleHandle(0);
	w.lpfnWndProc=&WndProc;
	w.lpszClassName=class_name;
	w.lpszMenuName=0;
	w.style=CS_VREDRAW|CS_HREDRAW;
	RegisterClassEx(&w);
}

int WINAPI WinMain(HINSTANCE hInstance,HINSTANCE hPrevInstance,LPSTR lpCmdLine,
	int nShowCmd)
{
	HANDLE hmutex;

	//Is this program already running?
	hmutex=OpenMutex(MUTEX_ALL_ACCESS,FALSE,mutex_name);
	if(hmutex) {
		CloseHandle(hmutex);
		MessageBox(0,"Only one instance of this program may be run at once.\n",
			"There can be only one",MB_OK|MB_ICONINFORMATION);
		return 0;
	}

	hmutex=CreateMutex(0,FALSE,mutex_name);

	//
	RegisterWindowClass();
	main_window=CreateWindow(class_name,"Uniformenu dummy window",
		WS_OVERLAPPED|WS_SYSMENU,CW_USEDEFAULT,CW_USEDEFAULT,CW_USEDEFAULT,
		CW_USEDEFAULT,0,0,GetModuleHandle(0),0);
#ifdef _DEBUG
	ShowWindow(main_window,nShowCmd);//so it may be closed manually.
#endif
	AddTrayIcon();
	SetTimer(main_window,timer_id,1000,0);
	for(;;) {
		int r;
		MSG msg;

		r=GetMessage(&msg,main_window,0,0);
		if(r==0||r==-1) {
			break;
		}
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
	KillTimer(main_window,timer_id);
	RemoveTrayIcon();
	DestroyWindow(main_window);
	ReleaseMutex(hmutex);
	CloseHandle(hmutex);
	free(windows);
	return 0;
}
